{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE TypeOperators #-}
{- ToDo:
type for submatrices, analogous to Vector.Slice
-}
module Numeric.BLAS.Matrix.RowMajor (
   Matrix,
   Square,
   Vector,
   height, width,
   reshapeHeight, reshapeWidth,
   mapHeight, mapWidth,
   -- ToDo: maybe define custom functions with type synonym Vector and Matrix?
   Array2.singleRow, Array2.flattenRow,
   Array2.singleColumn, Array2.flattenColumn,
   fromSquare,
   identity,
   diagonal,
   takeRow,
   takeColumn,
   Array2.toRowArray, fromRowArray,
   fromRows,
   fromBlockArray,
   fromNonEmptyBlockArray,
   above,
   beside,
   takeTop, takeBottom, takeVertCenter,
   takeLeft, takeRight, takeHorizCenter,
   transpose,
   tensorProduct,
   decomplex,
   recomplex,
   scaleRows,
   scaleColumns,
   scaleRowsReal,
   scaleColumnsReal,
   multiplyVectorLeft,
   multiplyVectorRight,
   multiply,
   sumRows,
   sumColumns,
   dotRowwise,
   innerRowwise,
   kronecker,
   ) where

import qualified Numeric.BLAS.Matrix.TransposablePrivate as Transp
import qualified Numeric.BLAS.Matrix.RowMajor.Square as Square
import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
import qualified Numeric.BLAS.Vector as Vector
import qualified Numeric.BLAS.Slice as Slice
import qualified Numeric.BLAS.Private as Private
import Numeric.BLAS.Matrix.TransposablePrivate
         (Transposable(Transposed, NonTransposed))
import Numeric.BLAS.Scalar (RealOf, zero, one)
import Numeric.BLAS.Private
         (ShapeInt, shapeInt, ComplexShape, ixReal, ixImaginary, pointerSeq)

import qualified Numeric.CBLAS.FFI.Extra as CBlas
import qualified Numeric.BLAS.FFI.Generic as Blas
import qualified Numeric.Netlib.Modifier as Modi
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class
import Numeric.Netlib.Modifier (Conjugation(NonConjugated,Conjugated))

import qualified Foreign.Marshal.Utils as Marshal
import Foreign.Marshal.Array (copyArray, advancePtr)
import Foreign.ForeignPtr (withForeignPtr, castForeignPtr)
import Foreign.Storable (Storable)

import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)
import Control.Monad (when)

import qualified Data.Array.Comfort.Boxed.Dim2 as BoxedArray2
import qualified Data.Array.Comfort.Boxed as BoxedArray
import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Storable.Dim2 as Array2
import qualified Data.Array.Comfort.Storable as CheckedArray
import qualified Data.Array.Comfort.Shape.SubSize as SubSize
import qualified Data.Array.Comfort.Shape.Unchecked as ShapeU
import qualified Data.Array.Comfort.Shape as Shape
import Data.Array.Comfort.Storable.Unchecked (Array(Array))
import Data.Array.Comfort.Shape ((::+)((::+)))

import qualified Data.Traversable as Trav
import qualified Data.Foldable as Fold
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.List.HT as ListHT
import Data.Map (Map)
import Data.Set (Set)
import Data.Foldable (forM_)
import Data.Complex (Complex)
import Data.Tuple.HT (mapPair, mapFst, mapSnd, double)


{- $setup
>>> :set -XTypeOperators
>>> import Test.NumberType_.Numeric.BLAS.Vector.Slice (forSliced2)
>>> import Test.NumberType_.Numeric.BLAS.Vector
>>>    (forVector, genVector, Number_, number_, real_)
>>> import Test.Slice (ShapeInt, shapeInt)
>>> -- import Test.Utility (approx)
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Block as BlockMatrix
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Square as Square
>>> import qualified Numeric.BLAS.Matrix.RowMajor as Matrix
>>> import qualified Numeric.BLAS.Vector.Slice as VectorSlice
>>> import qualified Numeric.BLAS.Vector as Vector
>>> import qualified Numeric.BLAS.Slice as Slice
>>> import qualified Numeric.Netlib.Class as Class
>>> import Numeric.BLAS.Matrix.RowMajor.Block ((&===), (&|||))
>>> import Numeric.BLAS.Vector.Slice ((|+|))
>>> import Numeric.BLAS.Scalar (RealOf)
>>> import Numeric.Netlib.Modifier (Conjugation(NonConjugated, Conjugated))
>>> import qualified Data.Array.Comfort.Storable.Dim2 as Array2
>>> import qualified Data.Array.Comfort.Storable as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import qualified Data.Array.Comfort.Boxed as BoxedArray
>>> import qualified Data.Foldable as Fold
>>> import qualified Data.Map as Map
>>> import qualified Data.Set as Set
>>> import qualified Data.Monoid.HT as Mn
>>> import Data.Array.Comfort.Shape ((::+)((::+)))
>>> import Data.Map (Map)
>>> import Data.Function.HT (Id)
>>> import qualified Test.QuickCheck as QC
>>> import Test.QuickCheck ((===))
>>>
>>> type Matrix = Matrix.Matrix ShapeInt ShapeInt
>>> type Real_ = RealOf Number_
>>>
>>> maxDim :: Int
>>> maxDim = 10
>>>
>>> forMatrix ::
>>>    (QC.Testable prop, QC.Arbitrary a, Class.Floating a, Show a) =>
>>>    QC.Gen a -> (Matrix a -> prop) -> QC.Property
>>> forMatrix genElem =
>>>    QC.forAllShrink
>>>       (do height <- fmap shapeInt $ QC.choose (0,maxDim)
>>>           width <- fmap shapeInt $ QC.choose (0,maxDim)
>>>           genVector (height, width) genElem)
>>>       (\a ->
>>>          let (height@(Shape.ZeroBased h), width@(Shape.ZeroBased w)) =
>>>                Array.shape a in
>>>          Mn.when (h > 0)
>>>             [let h2 = div h 2 in
>>>              Array2.takeTop $
>>>                Array.reshape
>>>                   (Shape.ZeroBased h2 ::+ Shape.ZeroBased (h-h2), width) a,
>>>              Array2.takeTop $
>>>                Array.reshape (Shape.ZeroBased (h-1) ::+ (), width) a]
>>>          ++
>>>          Mn.when (w > 0)
>>>             [let w2 = div w 2 in
>>>              Array2.takeLeft $
>>>                Array.reshape
>>>                   (height, Shape.ZeroBased w2 ::+ Shape.ZeroBased (w-w2)) a,
>>>              Array2.takeLeft $
>>>                Array.reshape (height, Shape.ZeroBased (w-1) ::+ ()) a])
>>>
>>> genSplitShape :: ShapeInt -> QC.Gen (ShapeInt ::+ ShapeInt)
>>> genSplitShape (Shape.ZeroBased n) = do
>>>    k <- QC.choose (0, n)
>>>    return $ Shape.ZeroBased k ::+ Shape.ZeroBased (n-k)
>>>
>>> genSplitShape3 :: ShapeInt -> QC.Gen (ShapeInt ::+ ShapeInt ::+ ShapeInt)
>>> genSplitShape3 (Shape.ZeroBased n) = do
>>>    k <- QC.choose (0, n)
>>>    j <- QC.choose (0, n)
>>>    let a = min k j; b = max k j
>>>    return $
>>>       Shape.ZeroBased a ::+ Shape.ZeroBased (b-a) ::+ Shape.ZeroBased (n-b)
-}


type Matrix height width = Array (height,width)
{- |
There is also 'Shape.Square'
but this would be incompatible with other matrix operations.
For definite squares see "Numeric.BLAS.Matrix.RowMajor.Square"
and for advanced type hacks you can use the @lapack@ package.
-}
type Square sh = Matrix sh sh
type Vector = Array


height :: Matrix height width a -> height
height = fst . Array.shape

width :: Matrix height width a -> width
width = snd . Array.shape

-- ToDo: maybe setHeight, newHeight?
-- ToDo: If width is zero, then changing height is allowed. I think we do not want this.
reshapeHeight ::
   (Shape.C height0, Shape.C height1, Shape.C width) =>
   height1 -> Matrix height0 width a -> Matrix height1 width a
reshapeHeight = CheckedArray.mapShape . mapFst . const

reshapeWidth ::
   (Shape.C height, Shape.C width0, Shape.C width1) =>
   width1 -> Matrix height width0 a -> Matrix height width1 a
reshapeWidth = CheckedArray.mapShape . mapSnd . const

mapHeight ::
   (Shape.C height0, Shape.C height1, Shape.C width) =>
   (height0 -> height1) -> Matrix height0 width a -> Matrix height1 width a
mapHeight = CheckedArray.mapShape . mapFst

mapWidth ::
   (Shape.C height, Shape.C width0, Shape.C width1) =>
   (width0 -> width1) -> Matrix height width0 a -> Matrix height width1 a
mapWidth = CheckedArray.mapShape . mapSnd



fromSquare :: Square.Matrix sh a -> Square sh a
fromSquare = Array.mapShape (double . Shape.squareSize)

{- |
>>> Matrix.identity (Shape.ZeroBased 0) :: Matrix.Square ShapeInt Real_
StorableArray.fromList (ZeroBased {... 0},ZeroBased {... 0}) []
>>> Matrix.identity (Shape.ZeroBased 3) :: Matrix.Square ShapeInt Real_
StorableArray.fromList (ZeroBased {... 3},ZeroBased {... 3}) [1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0]
-}
identity :: (Shape.C sh, Class.Floating a) => sh -> Square sh a
identity = fromSquare . Square.identity

diagonal ::
   (VectorSlice.C v, Shape.C sh, Class.Floating a) => v sh a -> Square sh a
diagonal = fromSquare . Square.diagonal


takeRow ::
   (Shape.Indexed height, Shape.C width, Shape.Index height ~ ix,
    Storable a) =>
   ix -> Matrix height width a -> Vector width a
takeRow ix (Array (height_,width_) x) =
   Array.unsafeCreateWithSize width_ $ \n yPtr ->
   withForeignPtr x $ \xPtr ->
      copyArray yPtr (advancePtr xPtr (n * Shape.offset height_ ix)) n

-- ToDo: can be implemented using 'slice'
takeColumn ::
   (Shape.C height, Shape.Indexed width, Shape.Index width ~ ix,
    Class.Floating a) =>
   ix -> Matrix height width a -> Vector height a
takeColumn ix (Array (height_,width_) x) =
   Array.unsafeCreateWithSize height_ $ \n yPtr -> evalContT $ do
      let offset = Shape.offset width_ ix
      nPtr <- Call.cint n
      xPtr <- ContT $ withForeignPtr x
      incxPtr <- Call.cint $ Shape.size width_
      incyPtr <- Call.cint 1
      liftIO $ Blas.copy nPtr (advancePtr xPtr offset) incxPtr yPtr incyPtr


{- |
prop> :{
   forMatrix number_ $ \a ->

   Matrix.transpose a
   ===
   Matrix.fromRowArray
      (Matrix.height a)
      (VectorSlice.slicesVector Slice.columnArray a)
:}
-}
fromRowArray ::
   (VectorSlice.C v) =>
   (Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   width -> BoxedArray.Array height (v width a) -> Matrix height width a
fromRowArray width_ arr =
   Array.reshape (BoxedArray.shape arr, width_) $
   fromRows width_ $ BoxedArray.toList arr

fromRows ::
   (VectorSlice.C v, Shape.C width, Eq width, Class.Floating a) =>
   width -> [v width a] -> Matrix ShapeInt width a
fromRows width_ rows =
   Array.unsafeCreateWithAutoSizes (shapeInt $ length rows, width_) $
         \(SubSize.Atom _, SubSize.Atom widthSize) dstPtr ->

      Marshal.with (fromIntegral widthSize) $ \nPtr ->
      forM_ (zip (pointerSeq widthSize dstPtr) rows) $
         \(yPtr, x) -> do
            Call.assert
               "Matrix.fromRows: non-matching vector size"
               (width_ == VectorSlice.shape x)
            evalContT $ do
               (xPtr,incxPtr) <- VectorSlice.sliceArg x
               incyPtr <- Call.cint 1
               liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr


{- |
Explicit parameters for the shape of the result matrix
allow for working with arrays of zero rows or columns.

>>> :{
   (id :: Id (array (height, Map Char ShapeInt) Number_)) $
   Matrix.fromBlockArray
      (Map.singleton 'A' (shapeInt 2) <> Map.singleton 'B' (shapeInt 3))
      Map.empty $
   BoxedArray.fromList (Set.fromList "AB", Set.empty) []
:}
StorableArray.fromList (fromList [('A',ZeroBased {... 2}),('B',ZeroBased {... 3})],fromList []) []

prop> :{
   forMatrix number_ $ \block ->
   let height = Map.singleton 'A' $ fst $ Array.shape block in
   let width  = Map.singleton '1' $ snd $ Array.shape block in

   Array.reshape (height,width) block
   ===
   Array2.fromBlockArray height width
      (BoxedArray.replicate (Set.singleton 'A', Set.singleton '1') block)
:}

prop> :{
   forMatrix number_ $ \blockA1 ->
   forMatrix number_ $ \blockB3 ->
   let shapeR0 = fst $ Array.shape blockA1 in
   let shapeR1 = fst $ Array.shape blockB3 in
   let shapeC0 = snd $ Array.shape blockA1 in
   QC.forAll (fmap shapeInt $ QC.choose (0,10)) $ \shapeC1 ->
   let shapeC2 = snd $ Array.shape blockB3 in
   QC.forAll (genVector (shapeR0, shapeC1) number_) $ \blockA2 ->
   QC.forAll (genVector (shapeR0, shapeC2) number_) $ \blockA3 ->
   QC.forAll (genVector (shapeR1, shapeC0) number_) $ \blockB1 ->
   QC.forAll (genVector (shapeR1, shapeC1) number_) $ \blockB2 ->

   let height = Map.fromList [('A',shapeR0),('B',shapeR1)] in
   let width  = Map.fromList [('1',shapeC0),('2',shapeC1),('3',shapeC2)] in

   Array.reshape (height,width)
      (BlockMatrix.toMatrix
         (blockA1 &||| Array2.beside blockA2 blockA3
          &===
          blockB1 &||| blockB2 &||| blockB3))
   ===
   Matrix.fromBlockArray height width
      (BoxedArray.fromList
         (Set.fromList "AB", Set.fromList "123")
         [blockA1, blockA2, blockA3,
          blockB1, blockB2, blockB3])
:}
-}
fromBlockArray ::
   (Ord row,    Shape.C height, Eq height) =>
   (Ord column, Shape.C width,  Eq width) =>
   (Class.Floating a) =>
   Map row height -> Map column width ->
   BoxedArray.Array (Set row, Set column) (Matrix height width a) ->
   Matrix (Map row height) (Map column width) a
fromBlockArray height_ width_ blocks =
   Array.unsafeCreate (height_, width_) $ \bPtr ->
   let (_, heightOffsets) = offsets $ BoxedArray.fromMap height_ in
   let (ldb,widthOffsets) = offsets $ BoxedArray.fromMap width_  in
   Fold.sequence_ $
   BoxedArray.zipWith
      (\((row,m,heighti), (col,n,widthi)) (Array shape a) ->
         if shape == (heighti,widthi)
            then withForeignPtr a $ \aPtr ->
                  CBlas.copyMatrix Modi.NonTransposed n m
                     aPtr n (advancePtr bPtr (row*ldb+col)) ldb
            else error "Matrix.RowMajor.fromBlockArray: block shape mismatch")
      (BoxedArray2.cartesian heightOffsets widthOffsets)
      blocks

offsets ::
   (Shape.C shape, Traversable f) => f shape -> (Int, f (Int, Int, shape))
offsets =
   Trav.mapAccumL
      (\pos shape -> let n = Shape.size shape in (pos+n, (pos,n,shape)))
      0

{- |
Only the outer @BoxedArray@ need to be non-empty.

>>> :{
   let shapeR0 = shapeInt 2; shapeR1 = shapeInt 3 in
   let shapeC0 = shapeInt 3; shapeC1 = shapeInt 2 in
   let block sh a = Array.replicate sh (a::Real_) in
   Matrix.fromBlockArray
      (Map.singleton 'A' shapeR0 <> Map.singleton 'B' shapeR1)
      (Map.singleton '1' shapeC0 <> Map.singleton '2' shapeC1) $
   BoxedArray.fromList (Set.fromList "AB", Set.fromList "12")
      [block (shapeR0,shapeC0) 0, block (shapeR0,shapeC1) 1,
       block (shapeR1,shapeC0) 2, block (shapeR1,shapeC1) 3]
:}
StorableArray.fromList (fromList [('A',ZeroBased {... 2}),('B',ZeroBased {... 3})],fromList [('1',ZeroBased {... 3}),('2',ZeroBased {... 2})]) [0.0,0.0,0.0,1.0,1.0,0.0,0.0,0.0,1.0,1.0,2.0,2.0,2.0,3.0,3.0,2.0,2.0,2.0,3.0,3.0,2.0,2.0,2.0,3.0,3.0]

prop> :{
   forMatrix number_ $ \blockA1 ->
   forMatrix number_ $ \blockB2 ->
   let shapeR0 = fst $ Array.shape blockA1 in
   let shapeC0 = snd $ Array.shape blockA1 in
   let shapeR1 = fst $ Array.shape blockB2 in
   let shapeC1 = snd $ Array.shape blockB2 in
   QC.forAll (genVector (shapeR0, shapeC1) number_) $ \blockA2 ->
   QC.forAll (genVector (shapeR1, shapeC0) number_) $ \blockB1 ->
   let blocked =
         BoxedArray.fromList (Set.fromList "AB", Set.fromList "12")
            [blockA1, blockA2, blockB1, blockB2] in

   Array2.fromNonEmptyBlockArray blocked
   ===
   Matrix.fromNonEmptyBlockArray blocked
:}

prop> :{
   forMatrix number_ $ \blockA1 ->
   forMatrix number_ $ \blockB2 ->
   forMatrix number_ $ \blockC3 ->
   let shapeR0 = fst $ Array.shape blockA1 in
   let shapeC0 = snd $ Array.shape blockA1 in
   let shapeR1 = fst $ Array.shape blockB2 in
   let shapeC1 = snd $ Array.shape blockB2 in
   let shapeR2 = fst $ Array.shape blockC3 in
   let shapeC2 = snd $ Array.shape blockC3 in
   QC.forAll (genVector (shapeR0, shapeC1) number_) $ \blockA2 ->
   QC.forAll (genVector (shapeR0, shapeC2) number_) $ \blockA3 ->
   QC.forAll (genVector (shapeR1, shapeC0) number_) $ \blockB1 ->
   QC.forAll (genVector (shapeR1, shapeC2) number_) $ \blockB3 ->
   QC.forAll (genVector (shapeR2, shapeC0) number_) $ \blockC1 ->
   QC.forAll (genVector (shapeR2, shapeC1) number_) $ \blockC2 ->
   let blocked =
         BoxedArray.fromList (Set.fromList "ABC", Set.fromList "123")
            [blockA1, blockA2, blockA3,
             blockB1, blockB2, blockB3,
             blockC1, blockC2, blockC3] in

   Array2.fromNonEmptyBlockArray blocked
   ===
   Matrix.fromNonEmptyBlockArray blocked
:}
-}
fromNonEmptyBlockArray ::
   (Ord row,    Shape.C height, Eq height) =>
   (Ord column, Shape.C width,  Eq width) =>
   (Class.Floating a) =>
   BoxedArray.Array (Set row, Set column) (Matrix height width a) ->
   Matrix (Map row height) (Map column width) a
fromNonEmptyBlockArray arr =
   let shapes = map Array.shape $ BoxedArray.toList arr in
   let w = Set.size $ snd $ BoxedArray.shape arr in
   let (rowIxs, columnIxs) =
         mapPair (Set.toAscList, Set.toAscList) $ BoxedArray.shape arr in
   case (ListHT.sieve w shapes, take w shapes) of
      (leftColumn@(_:_), topRow@(_:_)) ->
         fromBlockArray
            (Map.fromList $ zip rowIxs $ map fst leftColumn)
            (Map.fromList $ zip columnIxs $ map snd topRow)
            arr
      _ -> error "Matrix.RowMajor.fromNonEmptyBlockArray: empty array"


infixr 2 `above`
infixr 3 `beside`

{- |
prop> :{
   forMatrix number_ $ \a ->
   let (height, width) = Array.shape a in
   QC.forAll (genSplitShape height) $ \splitHeight ->
      let b = Array.reshape (splitHeight, width) a in
      b === Matrix.above (Matrix.takeTop b) (Matrix.takeBottom b)
:}
-}
above ::
   (Shape.C heightA, Shape.C heightB) =>
   (Shape.C width, Eq width) =>
   (Storable a) =>
   Matrix heightA width a ->
   Matrix heightB width a ->
   Matrix (heightA::+heightB) width a
above = Array2.above

{- |
prop> :{
   forMatrix number_ $ \a ->
   let (height, width) = Array.shape a in
   QC.forAll (genSplitShape width) $ \splitWidth ->
      let b = Array.reshape (height, splitWidth) a in
      b === Matrix.beside (Matrix.takeLeft b) (Matrix.takeRight b)
:}
-}
beside ::
   (Shape.C widthA, Shape.C widthB) =>
   (Shape.C height, Eq height) =>
   (Class.Floating a) =>
   Matrix height widthA a ->
   Matrix height widthB a ->
   Matrix height (widthA::+widthB) a
beside (Array (heightA,widthA) a) (Array (heightB,widthB) b) =
   Array.unsafeCreateWithAutoSizes (heightA,widthA::+widthB) $
      \(SubSize.Atom m,
         SubSize.Sub n (SubSize.Atom na ::+ SubSize.Atom nb)) cPtr ->
   withForeignPtr a $ \aPtr ->
   withForeignPtr b $ \bPtr -> do
   Call.assert "beside: heights mismatch" (heightA == heightB)
   CBlas.copyMatrix Modi.NonTransposed na m aPtr na cPtr n
   CBlas.copyMatrix Modi.NonTransposed nb m bPtr nb (advancePtr cPtr na) n

{-
prop> :{
   forMatrix number_ $ \a ->
   let (height, width) = Array.shape a in
   QC.forAll (genVector width number_) $ \x ->
   QC.forAll (genSplitShape height) $ \splitHeight ->
   let b = Array.reshape (splitHeight, width) a in

   Matrix.multiplyVectorRight b x
   ===
   Vector.append
      (Matrix.multiplyVectorRight (Matrix.takeTop b) x)
      (Matrix.multiplyVectorRight (Matrix.takeBottom b) x)
:}
-}
takeTop ::
   (Shape.C heightA, Shape.C heightB, Shape.C width, Storable a) =>
   Matrix (heightA::+heightB) width a ->
   Matrix heightA width a
takeTop = Array2.takeTop

takeBottom ::
   (Shape.C heightA, Shape.C heightB, Shape.C width, Storable a) =>
   Matrix (heightA::+heightB) width a ->
   Matrix heightB width a
takeBottom = Array2.takeBottom

takeVertCenter ::
   (Shape.C heightA, Shape.C heightB, Shape.C heightC, Shape.C width) =>
   (Storable a) =>
   Matrix (heightA::+heightB::+heightC) width a ->
   Matrix heightB width a
takeVertCenter = Array2.takeVertCenter

{- |
prop> :{
   forMatrix number_ $ \a ->
   let (height, width) = Array.shape a in
   QC.forAll (genVector width number_) $ \x ->
   QC.forAll (genSplitShape width) $ \splitWidth ->
   let b = Array.reshape (height, splitWidth) a in
   let y = Array.reshape splitWidth x in

   Matrix.multiplyVectorRight b y
   ===
   Vector.add
      (Matrix.multiplyVectorRight (Matrix.takeLeft b) (Vector.takeLeft y))
      (Matrix.multiplyVectorRight (Matrix.takeRight b) (Vector.takeRight y))
:}
-}
takeLeft ::
   (Shape.C height, Shape.C widthA, Shape.C widthB, Class.Floating a) =>
   Matrix height (widthA::+widthB) a ->
   Matrix height widthA a
takeLeft (Array (height_,widthA::+widthB) a) =
   Array.unsafeCreateWithAutoSizes (height_,widthA) $
      \(SubSize.Atom m, SubSize.Atom na) cPtr ->
   withForeignPtr a $ \aPtr -> do
   let nb = Shape.size widthB
   let n = na+nb
   CBlas.copyMatrix Modi.NonTransposed na m aPtr n cPtr na

takeRight ::
   (Shape.C height, Shape.C widthA, Shape.C widthB, Class.Floating a) =>
   Matrix height (widthA::+widthB) a ->
   Matrix height widthB a
takeRight (Array (height_,widthA::+widthB) a) =
   Array.unsafeCreateWithAutoSizes (height_,widthB) $
      \(SubSize.Atom m, SubSize.Atom nb) cPtr ->
   withForeignPtr a $ \aPtr -> do
   let na = Shape.size widthA
   let n = na+nb
   CBlas.copyMatrix Modi.NonTransposed nb m (advancePtr aPtr na) n cPtr nb



{- |
prop> :{
   forMatrix number_ $ \abc0 ->
   let width = Matrix.width abc0 in
   QC.forAll (genSplitShape3 width) $ \splitWidth ->
   let abc = Matrix.reshapeWidth splitWidth abc0 in

   let a = Matrix.takeLeft abc in
   let b = Matrix.takeHorizCenter abc in
   let c = Matrix.takeRight (Matrix.takeRight abc) in
   abc === BlockMatrix.toMatrix (a &||| b &||| c)
:}

prop> :{
   forMatrix number_ $ \abc0 ->
   let width = Matrix.width abc0 in
   QC.forAll (genSplitShape3 width) $ \splitWidth ->
   let abc = Matrix.reshapeWidth splitWidth abc0 in

   Matrix.takeHorizCenter abc
   ===
   Array2.takeHorizCenter abc
:}
-}
takeHorizCenter ::
   (Shape.C height, Shape.C widthA, Shape.C widthB, Shape.C widthC) =>
   (Class.Floating a) =>
   Matrix height (widthA::+widthB::+widthC) a ->
   Matrix height widthB a
takeHorizCenter (Array (height_,widthA::+widthB::+widthC) a) =
   Array.unsafeCreateWithAutoSizes (height_,widthB) $
      \(SubSize.Atom m, SubSize.Atom nb) dPtr ->
   withForeignPtr a $ \aPtr -> do
   let na = Shape.size widthA
   let nc = Shape.size widthC
   let n = na+nb+nc
   CBlas.copyMatrix Modi.NonTransposed nb m (advancePtr aPtr na) n dPtr nb



{- |
>>> Matrix.transpose (Array.fromList (Shape.Range 'a' 'c', shapeInt 2) [1,2,3,4,5,6::Real_])
StorableArray.fromList (ZeroBased {...2},Range {rangeFrom = 'a', rangeTo = 'c'}) [1.0,3.0,5.0,2.0,4.0,6.0]

prop> forMatrix number_ $ \a -> a == Matrix.transpose (Matrix.transpose a)

prop> :{
   forMatrix number_ $ \a ->

   Matrix.transpose a
   ==
   Array.reshape (Matrix.width a, Matrix.height a)
      (VectorSlice.concat
         (Fold.toList $ fmap VectorSlice.chunk $
          VectorSlice.slicesVector Slice.columnArray a))
:}

prop> :{
   forMatrix number_ $ \a ->

   fmap VectorSlice.toVector
      (VectorSlice.slicesVector Slice.rowArray (Matrix.transpose a))
   ==
   fmap VectorSlice.toVector
      (VectorSlice.slicesVector Slice.columnArray a)
:}
-}
transpose ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Matrix height width a -> Matrix width height a
transpose (Array (height_,width_) a) =
   Array.unsafeCreateWithAutoSizes (width_,height_) $
      \(SubSize.Atom n, SubSize.Atom m) bPtr ->
   withForeignPtr a $ \aPtr ->
   CBlas.copyMatrix Modi.Transposed n m aPtr n bPtr m



{- |
prop> :{
   forVector number_ $ \xs ->
   Matrix.tensorProduct (Left Conjugated) xs (Vector.one ())
   ===
   Matrix.singleColumn (Vector.conjugate xs)
:}

prop> :{
   forVector number_ $ \xs ->
   Matrix.tensorProduct (Right Conjugated) (Vector.one ()) xs
   ===
   Matrix.singleRow (Vector.conjugate xs)
:}

prop> :{
   forVector number_ $ \xs ->
   forVector number_ $ \ys ->
   QC.forAll (genVector (Array.shape ys) number_) $ \zs ->
   Matrix.multiplyVectorRight
      (Matrix.tensorProduct (Left NonConjugated) xs ys) zs
   ===
   Vector.scale (Vector.dot ys zs) xs
:}

prop> :{
   forVector number_ $ \xs ->
   forVector number_ $ \ys ->
   QC.forAll (genVector (Array.shape xs) number_) $ \zs ->
   Matrix.multiplyVectorLeft zs
      (Matrix.tensorProduct (Right NonConjugated) xs ys)
   ===
   Vector.scale (Vector.dot xs zs) ys
:}

prop> :{
   forVector number_ $ \xs ->
   forVector number_ $ \ys ->
   QC.forAll (genVector (Array.shape ys) number_) $ \zs ->
   Matrix.multiplyVectorRight
      (Matrix.tensorProduct (Right Conjugated) xs ys) zs
   ===
   Vector.scale (Vector.inner ys zs) xs
:}

prop> :{
   forVector number_ $ \xs ->
   forVector number_ $ \ys ->
   QC.forAll (genVector (Array.shape xs) number_) $ \zs ->
   Matrix.multiplyVectorLeft zs
      (Matrix.tensorProduct (Left Conjugated) xs ys)
   ===
   Vector.scale (Vector.inner xs zs) ys
:}
-}
-- ToDo: generalize to class that unifies Vector and VectorSlice Subvector
-- ToDo: there is also GER that allows for slices
tensorProduct ::
   (VectorSlice.Subvector v, VectorSlice.Subvector w) =>
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Either Conjugation Conjugation ->
   v height a -> w width a -> Matrix height width a
tensorProduct side x y =
   Array.unsafeCreateWithAutoSizes (VectorSlice.shape x, VectorSlice.shape y) $
      \(SubSize.Atom n, SubSize.Atom m) cPtr -> do
   let trans conjugated =
         case conjugated of NonConjugated -> 'T'; Conjugated -> 'C'
   let ((transa,transb),(lda,ldb)) =
         case side of
            Left c -> (('N', trans c),(m,n))
            Right c -> ((trans c, 'N'),(1,1))
   when (n>0 && m>0) $ evalContT $ do
      transaPtr <- Call.char transa
      transbPtr <- Call.char transb
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint 1
      alphaPtr <- Call.number one
      aPtr <- VectorSlice.startArg y
      ldaPtr <- Call.leadingDim lda
      bPtr <- VectorSlice.startArg x
      ldbPtr <- Call.leadingDim ldb
      betaPtr <- Call.number zero
      ldcPtr <- Call.leadingDim m
      liftIO $
         Blas.gemm
            transaPtr transbPtr mPtr nPtr kPtr
            alphaPtr aPtr ldaPtr bPtr ldbPtr betaPtr cPtr ldcPtr


decomplex ::
   (Class.Real a) =>
   Matrix height width (Complex a) ->
   Matrix height (width, ComplexShape) a
decomplex (Array (height_,width_) a) =
   Array (height_, (width_, Shape.static)) (castForeignPtr a)

recomplex ::
   (Class.Real a) =>
   Matrix height (width, ComplexShape) a ->
   Matrix height width (Complex a)
recomplex (Array (height_, (width_, Shape.NestedTuple _)) a) =
   Array (height_,width_) (castForeignPtr a)


{- |
prop> :{
   forMatrix number_ $ \a ->
   a === Matrix.scaleRows (VectorSlice.one (Matrix.height a)) a
:}

prop> :{
   forSliced2 number_ $ \x y ->
   QC.forAll (Shape.ZeroBased <$> QC.choose (1,maxDim)) $ \width ->
   QC.forAll (genVector (VectorSlice.shape x, width) number_) $ \a ->

   Matrix.scaleRows (x|+|y) a
   ===
   Matrix.scaleRows x a |+| Matrix.scaleRows y a
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (Shape.ZeroBased <$> QC.choose (1,maxDim)) $ \set ->
   QC.forAll (genVector (Matrix.height a, set) number_) $ \b ->
   QC.forAll (QC.elements $ Shape.indices set) $ \k ->

   Matrix.scaleRows (VectorSlice.sliceVector (Slice.column k) b) a
   ===
   Matrix.scaleRows (Matrix.takeColumn k b) a
:}
-}
scaleRows ::
   (VectorSlice.C v) =>
   (Shape.C height, Eq height, Shape.C width, Class.Floating a) =>
   v height a -> Matrix height width a -> Matrix height width a
scaleRows x (Array shape a) =
   Array.unsafeCreateWithAutoSizes shape $
      \(SubSize.Atom m, SubSize.Atom n) bPtr -> do

   Call.assert "scaleRows: sizes mismatch" (VectorSlice.shape x == fst shape)
   evalContT $ do
      nPtr <- Call.cint n
      xPtr <- VectorSlice.startArg x
      aPtr <- ContT $ withForeignPtr a
      incaPtr <- Call.cint 1
      incbPtr <- Call.cint 1
      liftIO $ sequence_ $ take m $
         zipWith3
            (\xkPtr akPtr bkPtr -> do
               Blas.copy nPtr akPtr incaPtr bkPtr incbPtr
               Blas.scal nPtr xkPtr bkPtr incbPtr)
            (pointerSeq (VectorSlice.increment x) xPtr)
            (pointerSeq n aPtr)
            (pointerSeq n bPtr)

{- |
prop> :{
   forMatrix number_ $ \a ->
   a === Matrix.scaleColumns (VectorSlice.one (Matrix.width a)) a
:}

prop> :{
   forSliced2 number_ $ \x y ->
   QC.forAll (Shape.ZeroBased <$> QC.choose (1,maxDim)) $ \height ->
   QC.forAll (genVector (height, VectorSlice.shape x) number_) $ \a ->

   Matrix.scaleColumns (x|+|y) a
   ===
   Matrix.scaleColumns x a |+| Matrix.scaleColumns y a
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (Shape.ZeroBased <$> QC.choose (1,maxDim)) $ \set ->
   QC.forAll (genVector (Matrix.width a, set) number_) $ \b ->
   QC.forAll (QC.elements $ Shape.indices set) $ \k ->

   Matrix.scaleColumns (VectorSlice.sliceVector (Slice.column k) b) a
   ===
   Matrix.scaleColumns (Matrix.takeColumn k b) a
:}
-}
scaleColumns ::
   (VectorSlice.C v) =>
   (Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   v width a -> Matrix height width a -> Matrix height width a
scaleColumns x (Array shape a) =
   Array.unsafeCreateWithAutoSizes shape $
      \(SubSize.Atom m, SubSize.Atom n) bPtr -> do

   Call.assert "scaleColumns: sizes mismatch" (VectorSlice.shape x == snd shape)
   evalContT $ do
      transPtr <- Call.char 'T'
      nPtr <- Call.cint n
      klPtr <- Call.cint 0
      kuPtr <- Call.cint 0
      alphaPtr <- Call.number one
      (xPtr,ldxPtr) <- VectorSlice.sliceFillArg x
      aPtr <- ContT $ withForeignPtr a
      incaPtr <- Call.cint 1
      betaPtr <- Call.number zero
      incbPtr <- Call.cint 1
      liftIO $ sequence_ $ take m $
         zipWith
            (\akPtr bkPtr ->
               Private.gbmv transPtr
                  nPtr nPtr klPtr kuPtr alphaPtr xPtr ldxPtr
                  akPtr incaPtr betaPtr bkPtr incbPtr)
            (pointerSeq n aPtr)
            (pointerSeq n bPtr)


newtype ScaleReal f g a = ScaleReal {getScaleReal :: f (RealOf a) -> g a -> g a}

{- |
prop> :{
   forMatrix number_ $ \a ->
   let (height, _width) = Array.shape a in
   QC.forAll (genVector height real_) $ \x ->

   Matrix.scaleRowsReal x a
   ===
   Matrix.scaleRows (Vector.fromReal x) a
:}
-}
scaleRowsReal ::
   (VectorSlice.C v) =>
   (Shape.C height, Eq height, Shape.C width, Class.Floating a) =>
   v height (RealOf a) -> Matrix height width a -> Matrix height width a
scaleRowsReal =
   getScaleReal $
   Class.switchFloating
      (ScaleReal scaleRows)
      (ScaleReal scaleRows)
      (ScaleReal $ \x -> recomplex . scaleRows x . decomplex)
      (ScaleReal $ \x -> recomplex . scaleRows x . decomplex)

{- |
prop> :{
   forMatrix number_ $ \a ->
   let (_height, width) = Array.shape a in
   QC.forAll (genVector width real_) $ \x ->

   Matrix.scaleColumnsReal x a
   ===
   Matrix.scaleColumns (Vector.fromReal x) a
:}
-}
scaleColumnsReal ::
   (VectorSlice.C v) =>
   (Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   v width (RealOf a) -> Matrix height width a -> Matrix height width a
scaleColumnsReal =
   getScaleReal $
   Class.switchFloating
      (ScaleReal scaleColumns)
      (ScaleReal scaleColumns)
      (ScaleReal $ \x -> recomplex . scaleColumns (dup x) . decomplex)
      (ScaleReal $ \x -> recomplex . scaleColumns (dup x) . decomplex)

{-
prop> :{
   \x ->
   dup x ==
   tensorProduct (Left NonConjugated) x (Vector.one Shape.NestedTuple)
:}
-}
dup ::
   (VectorSlice.C v) =>
   (Shape.C sh, Class.Floating a) =>
   v sh a -> Matrix sh ComplexShape a
dup x =
   Array.unsafeCreateWithAutoSizes (VectorSlice.shape x, Shape.static) $
      \(SubSize.Atom n, SubSize.Atom _m) yPtr -> evalContT $ do
   nPtr <- Call.cint n
   (xPtr, incxPtr) <- VectorSlice.sliceArg x
   incyPtr <- Call.cint 2
   liftIO $ do
      Blas.copy nPtr xPtr incxPtr yPtr incyPtr
      Blas.copy nPtr xPtr incxPtr (advancePtr yPtr 1) incyPtr


{- |
>>> Matrix.multiplyVectorLeft (Array.vectorFromList [3,1,4]) (Array.fromList (Shape.ZeroBased (3::Int), Shape.Range 'a' 'b') [0,1,0,0,1,0::Real_])
StorableArray.fromList (Range {rangeFrom = 'a', rangeTo = 'b'}) [4.0,3.0]

prop> :{
   forVector number_ $ \xs ->
   Matrix.multiplyVectorLeft xs (Matrix.identity (Array.shape xs)) == xs
:}
-}
multiplyVectorLeft ::
   (VectorSlice.C v) =>
   (Eq height, Shape.C height, Shape.C width, Class.Floating a) =>
   v height a -> Matrix height width a -> Vector width a
multiplyVectorLeft x a = Transp.multiplyVector x (NonTransposed a)

{- |
>>> Matrix.multiplyVectorRight (Array.fromList (Shape.Range 'a' 'b', Shape.ZeroBased (3::Int)) [0,0,1,1,0,0]) (Array.vectorFromList [3,1,4::Real_])
StorableArray.fromList (Range {rangeFrom = 'a', rangeTo = 'b'}) [4.0,3.0]
>>> Matrix.multiplyVectorRight (Array.fromList (Shape.Range 'a' 'b', Shape.ZeroBased (3::Int)) [2,7,1,8,2,8]) (Array.vectorFromList [3,1,4::Real_])
StorableArray.fromList (Range {rangeFrom = 'a', rangeTo = 'b'}) [17.0,58.0]

prop> :{
   forVector number_ $ \xs ->
   Matrix.multiplyVectorRight (Matrix.identity (Array.shape xs)) xs === xs
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (snd $ Array.shape a) number_) $ \x ->
   Matrix.singleColumn (Matrix.multiplyVectorRight a x)
   ===
   Matrix.multiply a (Matrix.singleColumn x)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (fst $ Array.shape a) number_) $ \x ->
   QC.forAll (genVector (snd $ Array.shape a) number_) $ \y ->
   Vector.dot x (Matrix.multiplyVectorRight a y)
   ===
   Vector.dot (Matrix.multiplyVectorLeft x a) y
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (snd $ Array.shape a) number_) $ \x ->
   Matrix.multiplyVectorRight a x
   ===
   Matrix.multiplyVectorLeft x (Matrix.transpose a)
:}
-}
multiplyVectorRight ::
   (VectorSlice.C v) =>
   (Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   Matrix height width a -> v width a -> Vector height a
multiplyVectorRight a x = Transp.multiplyVector x (Transposed a)


{- |
>>> :{
   Matrix.multiply
      (Array.fromList (shapeInt 2, shapeInt 2) [1000,100,10,1])
      (Array.fromList (shapeInt 2, shapeInt 3) [0..5::Real_])
:}
... [300.0,1400.0,2500.0,3.0,14.0,25.0]

prop> :{
   forMatrix number_ $ \a ->
      Matrix.multiply (Matrix.identity (Matrix.height a)) a == a
:}
prop> :{
   forMatrix number_ $ \a ->
      Matrix.multiply a (Matrix.identity (Matrix.width a)) == a
:}
prop> :{
   forMatrix number_ $ \a ->
   forMatrix number_ $ \c ->
   QC.forAll (genVector (Matrix.width a, Matrix.height c) number_) $ \b ->
      Matrix.multiply a (Matrix.multiply b c)
      ===
      Matrix.multiply (Matrix.multiply a b) c
:}
-}
multiply ::
   (Shape.C height, Shape.C width, Shape.C fuse, Eq fuse, Class.Floating a) =>
   Matrix height fuse a -> Matrix fuse width a -> Matrix height width a
multiply a b = Transp.multiply (NonTransposed a) (NonTransposed b)



newtype Sums f g a = Sums {getSums :: f a -> g a}


{- |
Sum up each row.

prop> :{
   forMatrix number_ $ \a ->

   Matrix.sumRows a
   ===
   Matrix.multiplyVectorRight a (Vector.one $ snd $ Array.shape a)
:}

prop> :{
   forMatrix number_ $ \a ->
   Vector.sum (Matrix.sumRows a)
   ===
   Vector.sum (Matrix.sumColumns a)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (fst $ Array.shape a) number_) $ \x ->

   Vector.dot x (Matrix.sumRows a)
   ===
   Vector.sum (Matrix.multiplyVectorLeft x a)
:}
-}
sumRows, sumRowsPlain ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Matrix height width a -> Vector height a
sumRows a =
   case Shape.size $ width a of
      0 -> Vector.zero $ height a
      1 -> Array.mapShape fst a
      _ ->
         getSums
            (Class.switchFloating
               (Sums sumRowsPlain)
               (Sums sumRowsPlain)
               (Sums sumRowsComplex)
               (Sums sumRowsComplex))
            a

sumRowsPlain a =
   multiplyVectorRight
      (mapWidth ShapeU.Unchecked a)
      (Vector.one $ ShapeU.Unchecked $ snd $ Array.shape a)
--   multiplyVectorRight a (CheckedArray.replicate (snd $ Array.shape a) one)

{-
This approach halves the number of needed multiplications
at the expense of temporary arrays and additional copying.
We could save a bit of allocation and copying
with a low-level implementation.
Eventually, this may not pay off.

Alternative: Perform a Matrix.transpose trick
we would need a batched transpose of type:
   Array (height, (width,c)) a -> Array ((height, c), width) a

Or simpler:
Perform (Complex valued) matrix transpose and use sumColumns.

A batched matrix multiplication like batched_gemm in MKL
would also solve the problem,
but it requires a Vector.one (height, width).
-}
sumRowsComplex ::
   (Shape.C height, Shape.C width, Class.Real a) =>
   Matrix height width (Complex a) -> Vector height (Complex a)
sumRowsComplex a =
   let ones = Vector.one $ ShapeU.Unchecked $ snd $ Array.shape a in
   let aa =
         Array.mapShape
            (\(h,(w,c)) -> ((ShapeU.Unchecked h, ShapeU.Unchecked w), c)) $
         decomplex a in
   Array.mapShape ShapeU.deconsUnchecked $
   Vector.zipComplex
      (multiplyVectorRight
         (VectorSlice.extract (Slice.column ixReal) aa) ones)
      (multiplyVectorRight
         (VectorSlice.extract (Slice.column ixImaginary) aa) ones)


{- |
Sum up each column.

prop> :{
   forMatrix number_ $ \a ->

   Matrix.sumColumns a
   ===
   Matrix.multiplyVectorLeft (Vector.one $ fst $ Array.shape a) a
:}

prop> :{
   forMatrix number_ $ \a ->

   Matrix.sumColumns a
   ===
   Matrix.sumRows (Matrix.transpose a)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (snd $ Array.shape a) number_) $ \x ->

   Vector.dot (Matrix.sumColumns a) x
   ===
   Vector.sum (Matrix.multiplyVectorRight a x)
:}
-}
sumColumns, sumColumnsPlain ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Matrix height width a -> Vector width a
sumColumns a =
   case Shape.size $ height a of
      0 -> Vector.zero $ width a
      1 -> Array.mapShape snd a
      _ ->
         getSums
            (Class.switchFloating
               (Sums sumColumnsPlain)
               (Sums sumColumnsPlain)
               (Sums sumColumnsComplex)
               (Sums sumColumnsComplex))
            a

sumColumnsPlain a =
   multiplyVectorLeft
      (Vector.one $ ShapeU.Unchecked $ fst $ Array.shape a)
      (mapHeight ShapeU.Unchecked a)
--   multiplyVectorLeft (Vector.one $ fst $ Array.shape a) a
--   multiplyVectorLeft (CheckedArray.replicate (fst $ Array.shape a) one) a

sumColumnsComplex ::
   (Shape.C height, Shape.C width, Class.Real a) =>
   Matrix height width (Complex a) -> Vector width (Complex a)
sumColumnsComplex = Vector.recomplex . sumColumnsPlain . decomplex



{- |
Compute dot products of corresponding matrix rows.

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Array.toBoxed (Matrix.dotRowwise a b)
   ===
   BoxedArray.zipWith Vector.dot (Matrix.toRowArray a) (Matrix.toRowArray b)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Square.takeDiagonal
      (Array.mapShape (Shape.Square . fst)
         (Matrix.multiply a (Matrix.transpose b)))
   ===
   Matrix.dotRowwise a b
:}

prop> :{
   forMatrix number_ $ \a ->
   let shape = Array.shape a in
   QC.forAll (genVector shape number_) $ \b ->

   Matrix.dotRowwise a b
   ===
   Matrix.sumRows (Vector.mul a b)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Vector.dot a b
   ===
   Vector.sum (Matrix.dotRowwise a b)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Matrix.dotRowwise a b
   ===
   Matrix.dotRowwise b a
:}
-}
dotRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Matrix height width a ->
   Matrix height width a ->
   Vector height a
dotRowwise a b = sumRows $ Vector.mul a b

{- |
Compute inner products of corresponding matrix rows.

prop> :{
   forMatrix number_ $ \a ->
   let shape = Array.shape a in
   QC.forAll (genVector shape number_) $ \b ->

   Matrix.innerRowwise a b
   ===
   Matrix.sumRows (Vector.mulConj a b)
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Matrix.innerRowwise a b
   ===
   Matrix.dotRowwise (Vector.conjugate a) b
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Vector.inner a b
   ===
   Vector.sum (Matrix.innerRowwise a b)
:}
-}
innerRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Matrix height width a ->
   Matrix height width a ->
   Vector height a
innerRowwise a b = sumRows $ Vector.mulConj a b


{- |
>>> :{
   Matrix.kronecker
      (Array.fromList (shapeInt 2, shapeInt 2) [0,1,-1,0::Real_])
      (Array.fromList (shapeInt 2, shapeInt 3) [1..6])
:}
... [0.0,0.0,0.0,1.0,2.0,3.0,0.0,0.0,0.0,4.0,5.0,6.0,-1.0,-2.0,-3.0,0.0,0.0,0.0,-4.0,-5.0,-6.0,0.0,0.0,0.0]

>>> :{
   Matrix.kronecker
      (Array.fromList (shapeInt 2, shapeInt 2) [1,2,3,4::Real_])
      (Array.fromList (shapeInt 2, shapeInt 3) [1,2,4,8,16,32])
:}
... [1.0,2.0,4.0,2.0,4.0,8.0,8.0,16.0,32.0,16.0,32.0,64.0,3.0,6.0,12.0,4.0,8.0,16.0,24.0,48.0,96.0,32.0,64.0,128.0]

prop> :{
   QC.forAll (QC.choose (0,5)) $ \m ->
   QC.forAll (QC.choose (0,5)) $ \n ->
      Matrix.kronecker
         (Matrix.identity (shapeInt m))
         (Matrix.identity (shapeInt n))
      ==
      (Matrix.identity (shapeInt m, shapeInt n)
         :: Matrix.Square (ShapeInt, ShapeInt) Number_)
:}
-}
kronecker ::
   (Shape.C heightA, Shape.C widthA, Shape.C heightB, Shape.C widthB,
    Class.Floating a) =>
   Matrix heightA widthA a ->
   Matrix heightB widthB a ->
   Matrix (heightA,heightB) (widthA,widthB) a
kronecker a b = Transp.kroneckerLeft (NonTransposed a) b
