module Numeric.BLAS.Matrix.TransposablePrivate where

import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
import Numeric.BLAS.Scalar (zero, one)
import Numeric.BLAS.Private (fill, pointerSeq, copyConjugate, clacgv, mul, gemv)

import qualified Numeric.BLAS.FFI.Generic as Blas
-- import qualified Numeric.BLAS.FFI.Complex as BlasComplex
import qualified Numeric.BLAS.FFI.Real as BlasReal
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class
import Numeric.Netlib.Modifier (Conjugation(NonConjugated, Conjugated))

import qualified Foreign.Marshal.Array.Guarded as ForeignArray
import Foreign.Marshal.Array (advancePtr)
import Foreign.ForeignPtr (ForeignPtr, withForeignPtr)
import Foreign.Ptr (Ptr)
import Foreign.Storable (Storable, poke)
import Foreign.C.Types (CInt)

import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)
import Control.Monad (when)
import Control.Applicative (liftA2)

import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Shape as Shape
import Data.Array.Comfort.Storable.Unchecked (Array(Array))

import Data.Foldable (for_)
import Data.Complex (Complex)
import Data.Tuple.HT (mapPair, swap)
import Data.Bool.HT (if')


{- $setup
>>> :set -XTypeOperators
>>> import Test.NumberType_.Numeric.BLAS.Matrix.RowMajor (forMatrix)
>>> import Test.NumberType_.Numeric.BLAS.Vector (genVector, number_)
>>> import Test.Slice (ShapeInt, shapeInt)
>>> import qualified Numeric.BLAS.Matrix.Transposable as Transp
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Square as Square
>>> import qualified Numeric.BLAS.Matrix.RowMajor as Matrix
>>> import qualified Numeric.BLAS.Vector as Vector
>>> import qualified Numeric.Netlib.Class as Class
>>> import Numeric.BLAS.Matrix.Transposable
>>>    (Transposable, transposed, nonTransposed)
>>> -- import Numeric.BLAS.Vector (Vector)
>>> -- import Numeric.Netlib.Modifier (Conjugation(NonConjugated, Conjugated))
>>> import qualified Numeric.Netlib.Modifier as Modi
>>> import qualified Data.Array.Comfort.Boxed as BoxedArray
>>> import qualified Data.Array.Comfort.Storable as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import Data.Tuple.HT (swap)
>>> import qualified Test.QuickCheck as QC
>>> import Test.QuickCheck ((===))
>>>
>>> maxDim :: Int
>>> maxDim = 10
>>>
>>> genIdentityTrans ::
>>>    (Shape.C sh, Class.Floating a) =>
>>>    sh -> QC.Gen (Transp.Transposable sh sh a)
>>> genIdentityTrans sh = do
>>>    trans <- QC.arbitrary
>>>    return $
>>>       case trans of
>>>          Modi.Transposed -> transposed (Matrix.identity sh)
>>>          Modi.NonTransposed -> nonTransposed (Matrix.identity sh)
>>>
>>> forTransposable ::
>>>    (QC.Testable prop, QC.Arbitrary a, Class.Floating a, Show a) =>
>>>    QC.Gen a -> (Transposable ShapeInt ShapeInt a -> prop) -> QC.Property
>>> forTransposable genElem prop =
>>>    forMatrix genElem $ \a trans -> prop $
>>>       case trans of
>>>          Modi.Transposed -> transposed a
>>>          Modi.NonTransposed -> nonTransposed a
>>>
>>> genTransposable ::
>>>    (Shape.C height, Shape.C width, Class.Floating a) =>
>>>    (height,width) -> QC.Gen a -> QC.Gen (Transposable height width a)
>>> genTransposable shape genElem =
>>>    QC.oneof
>>>       [fmap nonTransposed $ genVector shape genElem,
>>>        fmap transposed $ genVector (swap shape) genElem]
-}


type Matrix height width = Array (height,width)
type Vector = Array

{-
ToDo:
We could also define a Shape for this purpose.
-}
data Transposable height width a =
     NonTransposed (Matrix height width a)
   | Transposed (Matrix width height a)
   deriving (Show)


nonTransposed :: Matrix height width a -> Transposable height width a
nonTransposed = NonTransposed

transposed :: Matrix height width a -> Transposable width height a
transposed = Transposed


transpose :: Transposable height width a -> Transposable width height a
transpose at =
   case at of
      NonTransposed a -> Transposed a
      Transposed a -> NonTransposed a

mapShape ::
   (heightA -> heightB) ->
   (widthA -> widthB) ->
   Transposable heightA widthA a ->
   Transposable heightB widthB a
mapShape fHeight fWidth at =
   case at of
      NonTransposed a ->
         NonTransposed $ Array.mapShape (mapPair (fHeight, fWidth)) a
      Transposed a ->
         Transposed $ Array.mapShape (mapPair (fWidth, fHeight)) a


inspect :: Transposable height width a -> (Char, (height, width), ForeignPtr a)
inspect at =
   case at of
      NonTransposed (Array shA fptr) -> ('N', shA, fptr)
      Transposed (Array shA fptr) -> ('T', swap shA, fptr)


multiplyVector ::
   (VectorSlice.C v) =>
   (Shape.C height, Shape.C width, Eq height, Class.Floating a) =>
   v height a -> Transposable height width a -> Vector width a
multiplyVector x at =
   let (transChar, (height_,width_), a) = inspect at in
   Array.unsafeCreateWithSize width_ $ \m0 yPtr -> do
   Call.assert
      "Matrix.RowMajor.multiplyVector: shapes mismatch"
      (height_ == VectorSlice.shape x)
   let n0 = Shape.size height_
   let (m,n) =
         case at of
            NonTransposed _ -> (m0,n0)
            Transposed _ -> (n0,m0)
   if n0==0
      then fill zero m0 yPtr
      else evalContT $ do

      let lda = m
      transPtr <- Call.char transChar
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      alphaPtr <- Call.number one
      aPtr <- ContT $ withForeignPtr a
      ldaPtr <- Call.leadingDim lda
      (xPtr, incxPtr) <- VectorSlice.sliceFillArg x
      betaPtr <- Call.number zero
      incyPtr <- Call.cint 1
      liftIO $
         Blas.gemv
            transPtr mPtr nPtr alphaPtr aPtr ldaPtr
            xPtr incxPtr betaPtr yPtr incyPtr


{- |
prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genIdentityTrans (Matrix.height a)) $ \eye ->
      a === Transp.multiply eye (nonTransposed a)
:}
prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genIdentityTrans (Matrix.width a)) $ \eye ->
      a === Transp.multiply (nonTransposed a) eye
:}
prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genIdentityTrans (Matrix.width a)) $ \leftEye ->
   QC.forAll (genIdentityTrans (Matrix.height a)) $ \rightEye ->
      Transp.multiply leftEye (transposed a)
      ===
      Transp.multiply (transposed a) rightEye
:}
prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (QC.choose (0,maxDim)) $ \n ->
   QC.forAll (genVector (Matrix.width a, shapeInt n) number_) $ \b ->
      Matrix.transpose (Matrix.multiply a b)
      ===
      Transp.multiply (transposed b) (transposed a)
:}
-}
multiply ::
   (Shape.C height, Shape.C width, Shape.C fuse, Eq fuse, Class.Floating a) =>
   Transposable height fuse a ->
   Transposable fuse width a ->
   Matrix height width a
multiply a b = multiplyColumnMajor b a

multiplyColumnMajor ::
   (Shape.C height, Shape.C width, Shape.C fuse, Eq fuse, Class.Floating a) =>
   Transposable fuse height a ->
   Transposable width fuse a ->
   Matrix width height a
multiplyColumnMajor at bt =
   let (transa, (widthA,heightA), a) = inspect at in
   let (transb, (widthB,heightB), b) = inspect bt in
   Array.unsafeCreate (widthB,heightA) $ \cPtr -> do
   Call.assert
      "Matrix.RowMajor.multiply: shapes mismatch"
      (widthA == heightB)
   let m = Shape.size heightA
   let k = Shape.size widthA
   let n = Shape.size widthB
   let lda = case at of NonTransposed _ -> m; Transposed _ -> k
   let ldb = case bt of NonTransposed _ -> k; Transposed _ -> n
   let ldc = m

   when (n>0 && m>0) $ if' (k==0) (fill zero (m*n) cPtr) $ evalContT $ do
      transaPtr <- Call.char transa
      transbPtr <- Call.char transb
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint k
      alphaPtr <- Call.number one
      aPtr <- ContT $ withForeignPtr a
      ldaPtr <- Call.leadingDim lda
      bPtr <- ContT $ withForeignPtr b
      ldbPtr <- Call.leadingDim ldb
      betaPtr <- Call.number zero
      ldcPtr <- Call.leadingDim ldc
      liftIO $
         Blas.gemm
            transaPtr transbPtr mPtr nPtr kPtr alphaPtr aPtr ldaPtr
            bPtr ldbPtr betaPtr cPtr ldcPtr


{- ToDo:
use it in lapack:Test.Square.eigensystem and wuerfelschlange
See lapack.Test.Square.eigensystem - Square.takeDiagonal
also Test.Triangular.eigensystem
-}
{- |
Compute dot products of corresponding matrix rows.

prop> :{
   forTransposable number_ $ \a ->
   QC.forAll (genTransposable (Transp.shape a) number_) $ \b ->

   Array.toBoxed (Transp.dotRowwise a b)
   ===
   BoxedArray.zipWith Vector.dot
      (Matrix.toRowArray $ Transp.toMatrix a)
      (Matrix.toRowArray $ Transp.toMatrix b)
:}

prop> :{
   forTransposable number_ $ \a ->
   QC.forAll (genTransposable (Transp.shape a) number_) $ \b ->

   Square.takeDiagonal
      (Array.mapShape (Shape.Square . fst)
         (Transp.multiply a (Transp.transpose b)))
   ===
   Transp.dotRowwise a b
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Vector.dot a b
   ===
   Vector.sum (Transp.dotRowwise (transposed a) (transposed b))
:}

prop> :{
   forTransposable number_ $ \a ->
   QC.forAll (genTransposable (Transp.shape a) number_) $ \b ->

   Transp.dotRowwise a b
   ===
   Transp.dotRowwise b a
:}
-}
dotRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
dotRowwise =
   runRowwise $
   Class.switchFloating
      (Rowwise dotRowwiseRealOnce)
      (Rowwise dotRowwiseRealOnce)
      (Rowwise dotRowwiseComplex)
      (Rowwise dotRowwiseComplex)

{- |
Compute inner products of corresponding matrix rows.

prop> :{
   forTransposable number_ $ \a ->
   QC.forAll (genTransposable (Transp.shape a) number_) $ \b ->

   Transp.innerRowwise a b
   ===
   Transp.dotRowwise (Transp.conjugate a) b
:}

prop> :{
   forMatrix number_ $ \a ->
   QC.forAll (genVector (Array.shape a) number_) $ \b ->

   Vector.inner a b
   ===
   Vector.sum (Transp.innerRowwise (transposed a) (transposed b))
:}
-}
innerRowwise ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Floating a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
innerRowwise =
   runRowwise $
   Class.switchFloating
      (Rowwise dotRowwiseRealOnce)
      (Rowwise dotRowwiseRealOnce)
      (Rowwise innerRowwiseComplex)
      (Rowwise innerRowwiseComplex)

newtype Rowwise f g h a = Rowwise {runRowwise :: f a -> g a -> h a}



dotRowwiseRealOnce ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
dotRowwiseRealOnce at bt =
   case (at,bt) of
      (NonTransposed a, NonTransposed b) ->
         let (heightA,widthA) = Array.shape a in
         let (heightB,widthB) = Array.shape b in
         Array.unsafeCreateWithSize heightA $ \m dPtr -> do
         Call.assert "dotRowwise: heights mismatch" $ heightA == heightB
         Call.assert "dotRowwise: widths mismatch"  $ widthA  == widthB
         evalContT $ do
            let n = Shape.size widthA
            let nm = n*m
            aPtr <- ContT $ withForeignPtr $ Array.buffer a
            bPtr <- ContT $ withForeignPtr $ Array.buffer b
            cPtr <- ContT $ ForeignArray.alloca nm
            liftIO $ mul NonConjugated nm aPtr 1 bPtr 1 cPtr 1
--            liftIO $ do putStr "c = "; print . (id :: Integer -> Integer) . round =<< peek cPtr
            onesPtr <- ContT $ ForeignArray.alloca n
            liftIO $ fill one n onesPtr
            transPtr <- Call.char 'T'
            mPtr <- Call.cint m
            nPtr <- Call.cint n
            alphaPtr <- Call.number one
            ldcPtr <- Call.leadingDim n
            betaPtr <- Call.number zero
            incPtr <- Call.cint 1
            liftIO $
               gemv transPtr nPtr mPtr
                  alphaPtr cPtr ldcPtr onesPtr incPtr betaPtr dPtr incPtr
      (Transposed a, Transposed b) ->
--         let (heightA,widthA) = Transp.shape a in
--         let (heightB,widthB) = Transp.shape b in
         let (heightA,widthA) = swap $ Array.shape a in
         let (heightB,widthB) = swap $ Array.shape b in
         Array.unsafeCreateWithSize heightA $ \m dPtr -> do
         Call.assert "dotRowwise: heights mismatch" $ heightA == heightB
         Call.assert "dotRowwise: widths mismatch"  $ widthA  == widthB
         evalContT $ do
            let n = Shape.size widthA
            let nm = n*m
            aPtr <- ContT $ withForeignPtr $ Array.buffer a
            bPtr <- ContT $ withForeignPtr $ Array.buffer b
            cPtr <- ContT $ ForeignArray.alloca nm
            liftIO $ mul NonConjugated nm aPtr 1 bPtr 1 cPtr 1
            onesPtr <- ContT $ ForeignArray.alloca n
            liftIO $ fill one n onesPtr
            transPtr <- Call.char 'N'
            mPtr <- Call.cint m
            nPtr <- Call.cint n
            alphaPtr <- Call.number one
            ldcPtr <- Call.leadingDim m
            betaPtr <- Call.number zero
            incPtr <- Call.cint 1
            liftIO $
               gemv transPtr mPtr nPtr
                  alphaPtr cPtr ldcPtr onesPtr incPtr betaPtr dPtr incPtr
      _ -> dotRowwiseReal at bt



rowArgs ::
   (Shape.C height, Shape.C width, Storable a) =>
   Transposable height width a ->
   ((height, width), ContT r IO (Ptr a, Ptr CInt), Int)
rowArgs (NonTransposed x) =
   (Array.shape x,
      liftA2 (,) (ContT $ withForeignPtr $ Array.buffer x) (Call.cint 1),
      Shape.size $ snd $ Array.shape x)
rowArgs (Transposed x) =
   (swap $ Array.shape x,
      liftA2 (,)
         (ContT $ withForeignPtr $ Array.buffer x)
         (Call.cint $ Shape.size $ snd $ Array.shape x),
      1)

dotRowwiseReal ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Transposable height width a ->
   Transposable height width a ->
   Vector height a
dotRowwiseReal a b =
   let ((heightA,widthA), argsA, rowIncA) = rowArgs a in
   let ((heightB,widthB), argsB, rowIncB) = rowArgs b in
   Array.unsafeCreateWithSize heightA $ \m cPtr -> do
   Call.assert "dotRowwise: heights mismatch" $ heightA == heightB
   Call.assert "dotRowwise: widths mismatch"  $ widthA  == widthB
   evalContT $ do
      nPtr <- Call.cint $ Shape.size widthA
      (aPtr, incxPtr) <- argsA
      (bPtr, incyPtr) <- argsB
      liftIO $
         for_
            (take m $
               zip3
                  (pointerSeq rowIncA aPtr)
                  (pointerSeq rowIncB bPtr)
                  (pointerSeq 1 cPtr))
            $ \(xPtr,yPtr,zPtr) ->
               BlasReal.dot nPtr xPtr incxPtr yPtr incyPtr >>= poke zPtr

dotRowwiseComplex ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Transposable height width (Complex a) ->
   Transposable height width (Complex a) ->
   Vector height (Complex a)
dotRowwiseComplex a b =
   let ((heightA,widthA), argsA, rowIncA) = rowArgs a in
   let ((heightB,widthB), argsB, rowIncB) = rowArgs b in
   Array.unsafeCreateWithSize heightA $ \m cPtr -> do
   Call.assert "dotRowwise: heights mismatch" $ heightA == heightB
   Call.assert "dotRowwise: widths mismatch"  $ widthA  == widthB
   evalContT $ do
      transPtr <- Call.char 'N'
      nPtr <- Call.cint 1
      kPtr <- Call.cint $ Shape.size widthA
      alphaPtr <- Call.number one
      betaPtr <- Call.number zero
      (aPtr, ldaPtr)  <- argsA
      (bPtr, incyPtr) <- argsB
      inczPtr <- Call.cint 1
      liftIO $
         for_
            (take m $
               zip3
                  (pointerSeq rowIncA aPtr)
                  (pointerSeq rowIncB bPtr)
                  (pointerSeq 1 cPtr))
            $ \(xPtr,yPtr,zPtr) ->
               gemv
                  transPtr nPtr kPtr alphaPtr xPtr ldaPtr
                  yPtr incyPtr betaPtr zPtr inczPtr


innerRowwiseComplex ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Transposable height width (Complex a) ->
   Transposable height width (Complex a) ->
   Vector height (Complex a)
innerRowwiseComplex at bt =
   case (at,bt) of
      (NonTransposed a, _) ->
         innerRowsWithAnyComplex a bt NonConjugated
      (_, NonTransposed b) ->
         innerRowsWithAnyComplex b at Conjugated
      (Transposed a, Transposed _b) ->
         innerColumnsWithAnyComplex a bt

innerRowsWithAnyComplex ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Matrix height width (Complex a) ->
   Transposable height width (Complex a) ->
   Conjugation ->
   Vector height (Complex a)
innerRowsWithAnyComplex a b conj =
   let (heightA,widthA) = Array.shape a in
   let ((heightB,widthB), argsB, rowIncB) = rowArgs b in
   Array.unsafeCreateWithSize heightA $ \m cPtr -> do
   Call.assert "innerRowwise: heights mismatch" $ heightA == heightB
   Call.assert "innerRowwise: widths mismatch"  $ widthA  == widthB
   evalContT $ do
      transPtr <- Call.char 'C'
      let n = Shape.size widthA
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint 1
      alphaPtr <- Call.number one
      betaPtr <- Call.number zero
      aPtr <- ContT $ withForeignPtr $ Array.buffer a
      let rowIncA = n
      ldaPtr <- Call.leadingDim n
      (bPtr, incyPtr) <- argsB
      inczPtr <- Call.cint 1
      liftIO $ do
         for_
            (take m $
               zip3
                  (pointerSeq rowIncA aPtr)
                  (pointerSeq rowIncB bPtr)
                  (pointerSeq 1 cPtr))
            $ \(xPtr,yPtr,zPtr) ->
               gemv
                  transPtr nPtr kPtr alphaPtr xPtr ldaPtr
                  yPtr incyPtr betaPtr zPtr inczPtr
         case conj of
            NonConjugated -> return ()
            Conjugated -> clacgv mPtr cPtr inczPtr

{-
We could implement this without a temporary array
calling BlasReal.dot for all four combinations of real and imaginary parts.
But I am afraid this affects precision.
-}
innerColumnsWithAnyComplex ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Matrix width height (Complex a) ->
   Transposable height width (Complex a) ->
   Vector height (Complex a)
innerColumnsWithAnyComplex a b =
   let (widthA,heightA) = Array.shape a in
   let ((heightB,widthB), argsB, rowIncB) = rowArgs b in
   Array.unsafeCreateWithSize heightA $ \m cPtr -> do
   Call.assert "innerRowwise: heights mismatch" $ heightA == heightB
   Call.assert "innerRowwise: widths mismatch"  $ widthA  == widthB
   evalContT $ do
      transPtr <- Call.char 'C'
      let n = Shape.size widthA
      tempPtr <- Call.allocaArray n
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint 1
      alphaPtr <- Call.number one
      betaPtr <- Call.number zero
      aPtr <- ContT $ withForeignPtr $ Array.buffer a
      incxPtr <- Call.cint 1
      let rowIncA = 1
      ldaPtr <- Call.leadingDim n
      (bPtr, incyPtr) <- argsB
      inczPtr <- Call.cint 1
      liftIO $ do
         for_
            (take m $
               zip3
                  (pointerSeq rowIncA aPtr)
                  (pointerSeq rowIncB bPtr)
                  (pointerSeq 1 cPtr))
            $ \(xPtr,yPtr,zPtr) -> do
               Blas.copy nPtr xPtr mPtr tempPtr incxPtr
               gemv
                  transPtr nPtr kPtr alphaPtr tempPtr ldaPtr
                  yPtr incyPtr betaPtr zPtr inczPtr

-- not working correctly
_innerColumnsWithAnyComplex ::
   (Shape.C height, Shape.C width, Eq height, Eq width, Class.Real a) =>
   Matrix width height (Complex a) ->
   Transposable height width (Complex a) ->
   Vector height (Complex a)
_innerColumnsWithAnyComplex a b =
   let (widthA,heightA) = Array.shape a in
   let ((heightB,widthB), argsB, rowIncB) = rowArgs b in
   Array.unsafeCreateWithSize heightA $ \m cPtr -> do
   Call.assert "innerRowwise: heights mismatch" $ heightA == heightB
   Call.assert "innerRowwise: widths mismatch"  $ widthA  == widthB
   evalContT $ do
      transPtr <- Call.char 'N'
      let n = Shape.size widthA
      tempPtr <- Call.allocaArray n
      nPtr <- Call.cint n
      mPtr <- Call.cint m
      kPtr <- Call.cint 1
      alphaPtr <- Call.number one
      betaPtr <- Call.number zero
      aPtr <- ContT $ withForeignPtr $ Array.buffer a
      incxPtr <- Call.cint 1
      let rowIncA = 1
      ldaPtr <- Call.leadingDim n
      (bPtr, incyPtr) <- argsB
      inczPtr <- Call.cint 1
      liftIO $
         for_
            (take m $
               zip3
                  (pointerSeq rowIncA aPtr)
                  (pointerSeq rowIncB bPtr)
                  (pointerSeq 1 cPtr))
            $ \(xPtr,yPtr,zPtr) -> do
               copyConjugate nPtr xPtr mPtr tempPtr incxPtr
               gemv
                  transPtr kPtr nPtr alphaPtr tempPtr ldaPtr
                  yPtr incyPtr betaPtr zPtr inczPtr



kronecker ::
   (Shape.C heightA, Shape.C widthA, Shape.C heightB, Shape.C widthB,
    Class.Floating a) =>
   Transposable heightA widthA a ->
   Transposable heightB widthB a ->
   Transposable (heightA,heightB) (widthA,widthB) a
kronecker at bt =
   case bt of
      NonTransposed b -> NonTransposed $ kroneckerLeft at b
      Transposed b -> Transposed $ kroneckerLeft (transpose at) b

kroneckerLeft ::
   (Shape.C heightA, Shape.C widthA, Shape.C heightB, Shape.C widthB,
    Class.Floating a) =>
   Transposable heightA widthA a ->
   Matrix heightB widthB a ->
   Matrix (heightA,heightB) (widthA,widthB) a
kroneckerLeft at (Array (heightB,widthB) b) =
   let (_trans, (heightA,widthA), a) = inspect at
   in Array.unsafeCreate ((heightA,heightB), (widthA,widthB)) $ \cPtr ->
      evalContT $ do
   let (ma,na) = (Shape.size heightA, Shape.size widthA)
   let (mb,nb) = (Shape.size heightB, Shape.size widthB)
   let (lda,istep) =
         case at of
            NonTransposed _ -> (1,na)
            Transposed _ -> (ma,1)
   transaPtr <- Call.char 'N'
   transbPtr <- Call.char 'T'
   mPtr <- Call.cint na
   nPtr <- Call.cint nb
   kPtr <- Call.cint 1
   alphaPtr <- Call.number one
   aPtr <- ContT $ withForeignPtr a
   ldaPtr <- Call.leadingDim lda
   bPtr <- ContT $ withForeignPtr b
   ldbPtr <- Call.leadingDim 1
   betaPtr <- Call.number zero
   ldcPtr <- Call.leadingDim nb
   liftIO $ when (na>0 && nb>0) $
      for_ (liftA2 (,) (take ma [0..]) (take mb [0..])) $ \(i,j) -> do
         let aiPtr = advancePtr aPtr (istep*i)
         let bjPtr = advancePtr bPtr (nb*j)
         let cijPtr = advancePtr cPtr (na*nb*(j+mb*i))
         Blas.gemm
            transbPtr transaPtr nPtr mPtr kPtr alphaPtr
            bjPtr ldbPtr aiPtr ldaPtr betaPtr cijPtr ldcPtr
