{-# LANGUAGE TypeOperators #-}
module Numeric.BLAS.Subobject.View.Matrix where
{-
module Numeric.BLAS.Subobject.View.Matrix (
   T(..),
   module Numeric.BLAS.Subobject.View.Matrix,
   ) where
-}

import qualified Numeric.BLAS.Subobject.View as View
-- import qualified Numeric.BLAS.Subobject.Shape as Subshape
import qualified Numeric.BLAS.Subobject.Layout as Layout
-- import Numeric.BLAS.Subobject.View (T(..))
import Numeric.BLAS.Subobject.Layout (submatrixLeadingDim)

import qualified Data.Array.Comfort.Shape as Shape
import qualified Data.Array.Comfort.Boxed as BoxedArray
import Data.Array.Comfort.Shape ((::+)((::+)))

import qualified Data.List as List


{- $setup
>>> :set -XTypeOperators
>>>
>>> import qualified Numeric.BLAS.Subobject.View.Matrix as ViewMatrix
>>> import qualified Numeric.BLAS.Subobject.View as View
>>> import qualified Numeric.BLAS.Subobject.Layout as Layout
>>> import Test.Slice (shapeInt)
>>>
>>> import qualified Data.Array.Comfort.Boxed as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import Data.Array.Comfort.Shape ((::+)((::+)))
>>> import Data.Array.Comfort.Boxed ((!))
>>>
>>> import Control.Applicative (liftA3, pure)
>>>
>>> import qualified Test.QuickCheck as QC
>>>
>>> genView :: (height,width) -> QC.Gen (ViewMatrix.T (height,width))
>>> genView sh =
>>>    liftA3 View.Cons
>>>       (QC.choose (0,100))
>>>       (fmap Layout.Submatrix $ QC.choose (1,100))
>>>       (pure sh)
>>>
>>> type ShapeInt = Shape.ZeroBased Int
-}



type T = View.T Layout.Submatrix

fromShape ::
   (Shape.C height, Shape.C width) => (height,width) -> T (height,width)
fromShape sh@(_height,width) =
   View.Cons 0 (Layout.Submatrix (Shape.size width)) sh


reshapeHeight :: (Shape.C height0, Shape.C height1) =>
   height1 -> View.T lay (height0,width) -> View.T lay (height1,width)
reshapeHeight height1 (View.Cons s k (height0,width)) =
   View.Cons s k $
   if Shape.size height0 == Shape.size height1
      then (height1,width)
      else error "View.Matrix.reshapeHeight: shape sizes mismatch"

reshapeWidth :: (Shape.C width0, Shape.C width1) =>
   width1 -> View.T lay (height,width0) -> View.T lay (height,width1)
reshapeWidth width1 (View.Cons s k (height,width0)) =
   View.Cons s k $
   if Shape.size width0 == Shape.size width1
      then (height,width1)
      else error "View.Matrix.reshapeWidth: shape sizes mismatch"

mapShapeHeight :: (Shape.C height0, Shape.C height1) =>
   (height0 -> height1) ->
   View.T lay (height0,width) -> View.T lay (height1,width)
mapShapeHeight f view = reshapeHeight (f $ fst $ View.shape view) view

mapShapeWidth :: (Shape.C width0, Shape.C width1) =>
   (width0 -> width1) ->
   View.T lay (height,width0) -> View.T lay (height,width1)
mapShapeWidth f view = reshapeWidth (f $ snd $ View.shape view) view


row ::
   (Shape.Indexed height, Shape.C width) =>
   Shape.Index height -> T (height,width) -> View.Subvector width
row ix0 (View.Cons s (Layout.Submatrix ld) (height,width)) =
   View.Cons (s + Shape.offset height ix0 * ld) Layout.Subvector width

column ::
   (View.C lay, Shape.C height, Shape.Indexed width) =>
   Shape.Index width ->
   View.T lay (height,width) -> View.T Layout.Slice height
column ix1 (View.Cons s k (height,width)) =
   let (size1, offset1) = Shape.sizeOffset width
   in View.Cons
         (s + View.elemInc k * offset1 ix1)
         (Layout.Slice (View.rowOffset k $ Shape.ZeroBased size1))
         height

{- |
prop> :{
   QC.forAll (QC.choose (1,100)) $ \numRows ->
   QC.forAll (QC.choose (0,100)) $ \numColumns ->
   QC.forAll (genView (shapeInt numRows, shapeInt numColumns)) $ \view ->
   QC.forAll (QC.elements $ Shape.indices $ shapeInt numRows) $ \ix ->
   ViewMatrix.row ix view == ViewMatrix.rowArray view ! ix
:}
-}
rowArray ::
   (Shape.C height, Shape.C width) =>
   T (height,width) -> BoxedArray.Array height (View.Subvector width)
rowArray view =
   BoxedArray.fromList (fst $ View.shape view) $ rows view

rows ::
   (Shape.C height, Shape.C width) =>
   T (height,width) -> [View.Subvector width]
rows (View.Cons s (Layout.Submatrix ld) (sh0,sh1)) =
   List.map (\si -> View.Cons si Layout.Subvector sh1) $
   List.take (Shape.size sh0) $ iterate (ld+) s

{- |
prop> :{
   QC.forAll (QC.choose (0,100)) $ \numRows ->
   QC.forAll (QC.choose (1,100)) $ \numColumns ->
   QC.forAll (genView (shapeInt numRows, shapeInt numColumns)) $ \view ->
   QC.forAll (QC.elements $ Shape.indices $ shapeInt numColumns) $ \ix ->
   ViewMatrix.column ix view == ViewMatrix.columnArray view ! ix
:}
-}
columnArray ::
   (View.C lay, Shape.C height, Shape.C width) =>
   View.T lay (height,width) ->
   BoxedArray.Array width (View.T Layout.Slice height)
columnArray view =
   BoxedArray.fromList (snd $ View.shape view) $ columns view

columns ::
   (View.C lay, Shape.C height, Shape.C width) =>
   View.T lay (height,width) -> [View.T Layout.Slice height]
columns (View.Cons s k (sh0,sh1)) =
   let step = View.rowOffset k sh1
   in List.map (\si -> View.Cons si (Layout.Slice step) sh0) $
      List.take (Shape.size sh1) $ iterate (View.elemInc k +) s


{- |
prop> :{
   QC.forAll (QC.choose (1,100)) $ \numTopRows ->
   QC.forAll (QC.choose (1,100)) $ \numBottomRows ->
   QC.forAll (QC.choose (0,100)) $ \numColumns ->
   QC.forAll (genView (shapeInt numTopRows ::+ shapeInt numBottomRows,
                        shapeInt numColumns)) $ \view ->
   ViewMatrix.rowArray view
   ==
   Array.append
      (ViewMatrix.rowArray (ViewMatrix.top view))
      (ViewMatrix.rowArray (ViewMatrix.bottom view))
:}
-}
top ::
   (Shape.C height0, Shape.C height1, Shape.C width) =>
   T (height0::+height1,width) -> T (height0,width)
top (View.Cons s k (sh0::+_sh1, sh)) = View.Cons s k (sh0,sh)

bottom ::
   (Shape.C height0, Shape.C height1, Shape.C width) =>
   T (height0::+height1,width) -> T (height1,width)
bottom (View.Cons s k (sh0::+sh1, sh)) =
   View.Cons (s + Shape.size sh0 * submatrixLeadingDim k) k (sh1,sh)


{- |
prop> :{
   QC.forAll (QC.choose (1,100)) $ \numLeftColumns ->
   QC.forAll (QC.choose (1,100)) $ \numRightColumns ->
   QC.forAll (QC.choose (0,100)) $ \numRows ->
   QC.forAll
      (genView (shapeInt numRows,
                shapeInt numLeftColumns ::+ shapeInt numRightColumns)) $
      \view ->

   ViewMatrix.columnArray view
   ==
   Array.append
      (ViewMatrix.columnArray (ViewMatrix.left view))
      (ViewMatrix.columnArray (ViewMatrix.right view))
:}
-}
left ::
   (Shape.C height, Shape.C width0, Shape.C width1) =>
   T (height,width0::+width1) -> T (height,width0)
left (View.Cons s k (sh, sh0::+_sh1)) = View.Cons s k (sh,sh0)

right ::
   (Shape.C height, Shape.C width0, Shape.C width1) =>
   T (height,width0::+width1) -> T (height,width1)
right (View.Cons s k (sh, sh0::+sh1)) =
   View.Cons (s + Shape.size sh0) k (sh,sh1)



diagonal :: (Shape.C sh) =>
   T (Shape.Square sh) -> View.T Layout.Slice sh
diagonal (View.Cons s k (Shape.Square sh)) =
   View.Cons s (Layout.Slice $ submatrixLeadingDim k + 1) sh

cartesianFromSquare :: T (Shape.Square sh) -> T (sh,sh)
cartesianFromSquare = View.cartesianFromSquare

squareRow ::
   (Shape.Indexed sh) =>
   Shape.Index sh -> T (Shape.Square sh) -> View.Subvector sh
squareRow ix0 = row ix0 . cartesianFromSquare

squareColumn ::
   (View.C lay, Shape.Indexed sh) =>
   Shape.Index sh -> View.T lay (Shape.Square sh) -> View.T Layout.Slice sh
squareColumn ix1 = column ix1 . View.cartesianFromSquare
